import colour
import graphviz
from orglearn.mind_map.backend.backend import Backend


class Graphviz(Backend):
    def __init__(self, *args, **kwargs):
        self.ignore_shallow_tags = set(kwargs.get("ignore_shallow_tags_list", []))
        self.ignore_tags = set(kwargs.get("ignore_tags_list", []))
        self.base_color = colour.Color("green")

    def convert(self, tree, stream, **kwargs):
        # TODO: Maybe create heading from file name
        # self.dot = graphviz.Digraph(comment='asd')
        self.dot = graphviz.Digraph(comment="asd")
        # self.dot.attr(size='6,6')
        # self.dot.attr('graph', size='8.3,11.7!')
        # self.dot.attr('graph', size='11.7,8.3!')
        # self.dot.attr('graph', page='8.3,11.7!')
        # self.dot.attr('graph', page='11.7,8.3!')
        # self.dot.attr('graph', ratio='auto')
        self.dot.attr("graph", ratio="scale")
        self.dot.attr("graph", splines="spline")
        self.dot.attr("node", shape="box")
        self.dot.attr("graph", overlap="false")
        self.dot.attr("edge", arrowhead="vee", arrowtail="vee", arrowsize="0.75")
        # self.dot.attr('graph', mindist='5.0')
        # self.dot.engine = "neato"
        # self.dot.engine = "circo"
        self.dot.engine = "fdp"
        # self.dot.attr('graph', ratio='0.2')
        # self.dot.attr('graph', K='100')
        # self.dot.attr('graph', maxiter='100')
        try:
            # Try to set the center node text to a org file title comment
            tree.root.heading = tree._special_comments["TITLE"][0]
        except KeyError:
            tree.root.heading = "MAP"

        # Generate color gradient based on the depht of the org tree
        max_depth = 1
        for child in tree:
            max_depth = max(max_depth, child.level + 1)
        self.colors = list(self.base_color.range_to(colour.Color("white"), max_depth))

        self._process_node(tree.root)

        # TODO: Add option to split on highest level into files

        # TODO: Cannot take stream
        self.dot.render(stream.name)

    def _process_node(self, tree_node):
        """Create a map node from tree node and proccess its children."""

        # TODO(mato): What to do with a node body

        # First construct the current node
        # if tree_node.level == 0:
        #     self.dot.node(self._create_id(tree_node), tree_node.heading, shape='star', color='black')
        # elif tree_node.level == 1:
        #     self.dot.node(self._create_id(tree_node), tree_node.heading, shape='doublecircle')
        # else:
        # if tree_node.level == 0:
        #     self.dot.attr('node', shape='diamond', style='filled', color='lightgray')
        # else:
        #     self.dot.attr('node', shape='ellipse', color='black')
        # height: 0.5
        # width: 0.75
        scale = 0.80 ** tree_node.level
        height = str(1 * scale)
        width = str(2 * scale)
        self.dot.attr(
            "node",
            height=height,
            width=width,
            style="filled",
            fillcolor=self.colors[tree_node.level].get_hex_l(),
        )
        self.dot.node(self._create_id(tree_node), tree_node.heading)

        # If node has a parrent, create a link to it
        if tree_node.parent is not None:
            self.dot.edge(
                self._create_id(tree_node.parent), self._create_id(tree_node)
            )  # , constraint='false')

        # Process all children of this node
        for c in tree_node.children:
            if not self.ignore_tags.intersection(c.tags):
                self._process_node(c)

    def _create_id(self, node):
        """Hash the node to create identifier to reference nodes."""
        # TODO: We should double escape the '\' characters
        try:
            return (
                self._normalize_heading(node.parent.heading)
                + "%"
                + str(node.level)
                + "%"
                + self._normalize_heading(node.heading)
            )
        except AttributeError:
            return str(node.level) + "%" + self._normalize_heading(node.heading)

    def _normalize_heading(self, heading):
        """Normalize heading for dot format. Essentialy remove all ':' from headings."""
        return heading.replace(":", "")

    def get_ext(self):
        # Graphviz automatically appends the '.pdf'
        # And we don't want to colide with `pdf` command so prepend the '-map'
        # This results in: "<filename>-map.pdf"
        return "-map"



import pathlib

import pytest

from orglearn.preprocessor import Preprocessor


@pytest.fixture
def preprocessor():
    yield Preprocessor()


@pytest.fixture
def data_folder():
    yield pathlib.Path("./tests/data")



import contextlib
import logging
import os
import pathlib
import sys
import tempfile
import typing

from anki.collection import Collection


@contextlib.contextmanager
def Workdir(workdir: typing.Union[str, pathlib.Path]) -> typing.Iterator[None]:
    old_dir = os.getcwd()
    os.chdir(str(workdir))

    yield

    os.chdir(old_dir)


@contextlib.contextmanager
def create_empty_anki_collection() -> typing.Iterator[Collection]:
    """ContextManager used to create an empty Anki Collection.

    This collection is used for importing and exporting decks. For some reason
    creating a collection also changes the current working directory. Therefore
    it is necessary to call `close` as a cleanup."""

    (fd, path) = tempfile.mkstemp(suffix=".anki2", prefix="orglearn")
    try:
        os.close(fd)
        os.unlink(path)
    except OSError:
        pass

    col = Collection(path)

    yield col

    col.close(downgrade=False)


def setup_loggging(log_level: int) -> None:
    """Setup the default application logging."""

    # Configure the application logging
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    # Default logging is to stderr
    logging_handler = logging.StreamHandler(stream=sys.stderr)

    logging_handler.setFormatter(
        logging.Formatter(
            "[{asctime}] [{name:>15}:{lineno:<3}] [{levelname:.4}] -- {message}",
            style="{",
        )
    )

    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)
        handler.close()

    root_logger.addHandler(logging_handler)



from orglearn.mind_map.backend.graphviz import Graphviz


class Backends:

    BACKENDS = {}
    INSTANCES = {}

    @staticmethod
    def register_backend(b, s=None):
        # TODO: The name of the class is incorent and alway abcmeta
        Backends.BACKENDS[s or b.__class__.__name__.lower()] = b

    @staticmethod
    def get_backends():
        return Backends.BACKENDS.keys()

    @staticmethod
    def get_default_backend():
        return "graphviz"

    @staticmethod
    def make(backend_string, *args, **kwargs):

        try:
            return Backends.INSTANCES[backend_string]
        except KeyError:
            return Backends._instantiate(backend_string, *args, **kwargs)

    @staticmethod
    def _instantiate(backend_string, *args, **kwargs):
        assert backend_string in Backends.BACKENDS.keys()
        Backends.INSTANCES[backend_string] = Backends.BACKENDS[backend_string](
            *args, **kwargs
        )
        return Backends.INSTANCES[backend_string]


# Register all backends here
Backends.register_backend(Graphviz, "graphviz")



import enum
import logging
import pathlib
import random
import re
import typing

import genanki
import orgparse

log = logging.getLogger(__name__)

block_begin = re.compile(r"#\+BEGIN_(?P<block>[A-Z]+)( \w*)?")
block_end = re.compile(r"#\+END_(?P<block>[A-Z]+)")
latex_eq = re.compile(r"\$\$?(.*?)\$\$?")
image_struct = re.compile(r"\[\[(.*?)\]\]")
table = re.compile(r"^\|.*\|$")


class AnkiConvertMode(enum.Enum):
    """Mode of conversion."""

    NORMAL = enum.auto()
    BRIEF = enum.auto()
    CODE = enum.auto()


class NodeConvertor:
    """Convert orglearn node to anki Note."""

    MODEL_NORMAL = genanki.Model(
        # TODO: Abstract the id
        random.randrange(1 << 30, 1 << 31),
        "Orglearn - normal",
        fields=[{"name": "AField"}, {"name": "BField"}],
        templates=[
            {
                "name": "card1",
                "qfmt": "{{AField}}",
                "afmt": "{{FrontSide}}" '<hr id="answer">' "{{BField}}",
            }
        ],
        css=".card {text-align: left;}",
    )

    MODEL_CODE = genanki.Model(
        # TODO: Abstract the id
        random.randrange(1 << 30, 1 << 31),
        "Orglearn - code",
        fields=[{"name": "Task"}, {"name": "Assignment"}, {"name": "Solution"}],
        templates=[
            {
                "name": "card1",
                "qfmt": "{{Task}}" "</br>" "{{Assignment}}",
                "afmt": "{{FrontSide}}" '<hr id="answer">' "{{Solution}}",
            }
        ],
        css=".card {text-align: left;}",
    )

    def __init__(self) -> None:
        self._mode_convertors = {
            AnkiConvertMode.NORMAL: self._convert_normal,
            AnkiConvertMode.BRIEF: self._convert_brief,
            AnkiConvertMode.CODE: self._convert_code,
        }

        self.media_files: typing.List[str] = []

    def _get_card_title(self, node: orgparse.node.OrgNode) -> str:
        """Construct the node title."""
        res = node.heading
        while 1:
            try:
                node = node.parent
                res = f"{node.heading} -> {res}"
            except AttributeError:
                return res
        return res

    def convert(
        self, node: orgparse.node.OrgNode, mode: AnkiConvertMode = AnkiConvertMode.NORMAL
    ) -> typing.Optional[genanki.Note]:
        """Process a single node and append a new Note."""
        try:
            return self._mode_convertors[mode](node)
        except KeyError:
            log.error("Invalid convertor mode selected")
            raise ValueError("Invalid convertor mode")

    def _convert_normal(self, node: orgparse.node.OrgNode) -> typing.Optional[genanki.Note]:
        generate = False
        card_body = ""
        node_body = "\n".join(node._lines[1:])
        if node_body or not node.children:
            generate = True
            card_body = self._convert_text_to_anki(node_body)

        if "anki_list" in node.shallow_tags:
            generate = True
            card_body = self._append_anki_list_footer(node, card_body)

        card_body = card_body.replace("\n", "<br />")
        card_title = self._get_card_title(node)
        if generate:
            return genanki.Note(
                model=self.MODEL_NORMAL,
                fields=[card_title, card_body],
                tags=self.get_tags_from_node(node),
            )

        return None

    def _convert_brief(self, node: orgparse.node.OrgNode) -> typing.Optional[genanki.Note]:
        return None

    def _convert_code(self, node: orgparse.node.OrgNode) -> typing.Optional[genanki.Note]:
        generate = False
        assignment = ""
        solution = node.body
        node_body = "\n".join(node._lines[1:])
        if node_body or not node.children:
            generate = True

            # Split the node body into two parts beaking on the first 2 empty lines
            body_split = node_body.split("\n\n\n", 1)
            if len(body_split) == 2:
                assignment, solution = body_split

            assignment = self._convert_text_to_anki(assignment)
            solution = self._convert_text_to_anki(solution)

        if "anki_list" in node.shallow_tags:
            generate = True
            solution = self._append_anki_list_footer(node, solution)

        assignment = assignment.replace("\n", "<br />")
        solution = solution.replace("\n", "<br />")
        card_title = self._get_card_title(node)
        if generate:
            return genanki.Note(
                model=self.MODEL_CODE,
                fields=[card_title, assignment, solution],
                tags=self.get_tags_from_node(node),
            )

        return None

    def get_tags_from_node(self, node: orgparse.node.OrgNode) -> typing.List[str]:
        """Extract the list of tags from ORG."""
        return list(node.shallow_tags)

    def _append_anki_list_footer(self, node: orgparse.node.OrgNode, body: str) -> str:
        """Generate footer for the nodes marked with :anki_list:.

        This footer will be a list of child nodes. The rationale is to provide aditional
        context for the nodes that are the list of some topics.
        """
        if body:
            body += "\n"

        for child in node.children:
            body += "- {}\n".format(child.heading)
        return body

    def _convert_text_to_anki(self, body: str) -> str:
        """Perform necessary adjustments to the card text."""
        res = ""
        for line in body.splitlines(keepends=True):

            # LaTeX
            line = latex_eq.sub(r"[$]\1[/$]", line)

            # Block begin
            line = block_begin.sub(r"<pre><code>", line)

            # Block end
            line = block_end.sub(r"</code></pre>", line)

            # Tables
            match = table.match(line)
            if match:
                line = f"<pre><code>{line}</code></pre>"

            # Images
            match = image_struct.match(line)
            if match:
                line = image_struct.sub(
                    lambda x: f'<img src="{pathlib.Path(x.group(1)).name}">', line
                )
                # Collect the media files for exporting
                self.media_files.append(match.group(1))

            res += line

        return res



import orgparse


def test_no_preprocessing(preprocessor, data_folder):
    f = data_folder / "lokalne_prehladavanie.org"
    a = preprocessor.preprocess_file(str(f))
    b = f.read_text()

    assert a == b


def test_command_oi(preprocessor, data_folder):
    f = data_folder / "oi.org"
    a = preprocessor.preprocess_file(str(f))
    b = orgparse.loads(a)

    expected = "\nThis tests the command OI\n\n\nThis should be duplicated\n\n"
    assert b.children[0].body == expected

    expected = "\n\nmore body\n"
    assert b.children[2].body == expected

    assert b.children[3].body == "\n\ntest body\n"
    assert b.children[3].children[0].heading == "Nested"
    assert b.children[3].children[0].children[0].heading == "More"


def test_command_ois(preprocessor, data_folder):
    f = data_folder / "ois.org"
    a = preprocessor.preprocess_file(str(f))

    expected = """
* Main file

This tests the command OI


This should be duplicated


* Other heading

** Include me

This should be duplicated

* Testing header


more body

* Last header


test body

"""

    assert a == expected


def test_command_ol(preprocessor, data_folder):
    f = data_folder / "ol.org"
    a = preprocessor.preprocess_file(str(f))

    expected = """
* Main file

This tests the command OI

** [OL]Include me

This should be duplicated


* Other heading

** Include me

This should be duplicated

* Testing header

** [OL]More@./include.org

more body

* Last header

** [OL]Test@./include.org

test body

*** Nested

nested body

**** More

more body
"""
    assert a == expected


def test_invalid_command(preprocessor, data_folder):
    f = data_folder / "invalid_command.org"
    a = preprocessor.preprocess_file(str(f))

    expected = """
* Main file

This tests the command OI

** [ASD]Include me

* Other heading

** Include me

This should be duplicated

* Testing header

** [OL]More@./include.org

more body

* Last header

** [OL]Test@./include.org

test body

*** Nested

nested body

**** More

more body
"""

    assert a == expected



import os

import orgparse
from orglearn.preprocessor import Preprocessor


class MapConvertor:
    def __init__(self, o_file, f_list, backend, **kwargs):

        # TODO(mato): This method is duplicate
        # Try to deteremine output file if none was specified
        if o_file is None:
            root, ext = os.path.splitext(f_list[0])
            o_file = root + backend.get_ext()

        # Parse the org files into 1 tree
        # TODO(mato): For now we are expecting only one file, make it more generic
        for f in f_list[:1]:
            source = Preprocessor().preprocess_file(f)
            tree = orgparse.loads(source)

            with open(o_file, "w") as o_stream:
                backend.convert(tree, o_stream, **kwargs)



import logging
import pathlib
import re
import sys
import typing

import orgparse

import orglearn.utils as utils

# TODO: We should cache parsed files

log = logging.getLogger(__name__)


class Preprocessor:
    REGEXP_IMAGE = re.compile(r"^\s*\[\[(.*)\]\]$")
    REGEXP_COMMAND = re.compile(r"^(?P<stars>\*+)\s+\[(?P<command>[A-Z]+)\](?P<args>.*)$")

    def __init__(self) -> None:
        self.source_file: typing.Optional[str] = None

    def preprocess_file(self, file_path: str) -> str:
        """Preprocess file specified by the path."""
        log.info(f"Preprocessing file {file_path}")

        f = pathlib.Path(file_path)

        with utils.Workdir(f.parent):
            self.origin_file = pathlib.Path(f.name)
            self.current_file = file_path
            return self.preprocess_string(self.origin_file.read_text())

    @staticmethod
    def _find_node_in_tree(
        node_id: str, tree: orgparse.node.OrgNode
    ) -> typing.Optional[orgparse.node.OrgNode]:
        """Find node in the orgfile based on the node_id.

        Return `None` if node is not found or the node is not unique. If
        the `node_id` starts with '&' match on the 'ID' org property.
        """
        ret_node: orgparse.node.OrgNode = None
        node_id = node_id.strip()

        for node in tree:
            if node_id.startswith("&"):
                # Check the reference
                try:
                    # Strip the '&' prefix
                    match_string = node_id[1:]

                    if node.get_property("ID") == match_string:
                        if ret_node is not None:
                            log.warning(f"Node is not unique {node_id}")
                            return None
                        else:
                            ret_node = node
                except AttributeError:
                    # Root node does not have heading attribute, ignore it
                    pass
            else:
                try:
                    if node.heading == node_id:
                        if ret_node is not None:
                            log.warning(f"Node is not unique {node_id}")
                            return None
                        else:
                            ret_node = node
                except AttributeError:
                    # Root node does not have heading attribute, ignore it
                    pass

        return ret_node

    def preprocess_string(self, file_source: str) -> str:
        res = ""

        for line_num, line in enumerate(file_source.splitlines(keepends=True), start=1):
            m = self.REGEXP_COMMAND.search(line)

            if m:
                level = len(m.group("stars"))
                command = m.group("command")

                parts = m.group("args").split("@")
                if len(parts) == 1:
                    include_title = parts[0]
                    include_path = self.origin_file
                else:
                    include_title = parts[0]
                    include_path = parts[1] or self.origin_file

                include_org_file = orgparse.load(str(include_path))
                self.current_file = include_path

                if command == "OL":
                    res += line

                    node = self._find_node_in_tree(include_title, include_org_file)
                    if node is None:
                        self._abort_preprocessing(include_title, line_num)
                        return ""

                    res += self._process_body(node._lines[1:])
                    for child in node.children:
                        res += self._include_node(level + 1, child)
                elif command == "OI":

                    node = self._find_node_in_tree(include_title, include_org_file)
                    if node is None:
                        self._abort_preprocessing(include_title, line_num)
                        return ""

                    res += self._process_body(node._lines[1:])
                    for child in node.children:
                        res += self._include_node(level, child)
                elif command == "OIS":

                    node = self._find_node_in_tree(include_title, include_org_file)
                    if node is None:
                        self._abort_preprocessing(include_title, line_num)
                        return ""

                    res += self._process_body(node._lines[1:])
                else:
                    # This branch is here in case any of the titles start with [TAG] prefix
                    res += line
            else:
                res += line

        return res

    def _process_body(self, body: typing.List[str]) -> str:
        res = ""
        for line in body:
            m = self.REGEXP_IMAGE.search(line)

            if m:
                processed_file_base_dir = pathlib.Path(self.current_file).parent
                res_image_path = processed_file_base_dir / m.group(1)
                res_image_path = res_image_path.resolve(strict=True)

                if res_image_path.suffix not in [".png", ".jpg"]:
                    log.warning(
                        f"Image {res_image_path} might not be supported (consider converting to .png)"
                    )

                res += f"[[{str(res_image_path)}]]"
                res += "\n"
            else:
                res += line
                res += "\n"

        return res

    def _include_node(self, base_level: int, node: orgparse.node.OrgNode) -> str:
        res = ""

        # Get the heading line, this is a bit hackish, maybe a better way
        # is to construct the heading based on the public attributes
        heading = node._lines[0]

        # Strip the heading level and construct a new one
        heading = heading.lstrip("*")

        # Construct the correct heading
        res += "*" * base_level
        res += heading

        res += "\n"

        res += self._process_body(node._lines[1:])

        for child in node.children:
            res += self._include_node(base_level + 1, child)

        return res

    def _abort_preprocessing(self, node_id: str, line_num: int) -> None:
        # TODO: Replace this function with exception
        print(
            'Preprocessing file "{}" failed when searching for node_id "{}" in file "{}" referenced from "{}":{}'.format(
                self.origin_file, node_id, self.current_file, self.origin_file, line_num
            )
        )
        sys.exit(1)



import logging
import pathlib
import random
import typing

import genanki
import orgparse
from anki.exporting import AnkiPackageExporter
from anki.importing import AnkiPackageImporter

import orglearn.utils as utils
from orglearn.anki.node_convertor import AnkiConvertMode, NodeConvertor
from orglearn.preprocessor import Preprocessor

log = logging.getLogger(__name__)


class AnkiConvertor:
    """Org to Anki convertor."""

    ANKI_EXT = ".apkg"

    COMMENT_ANKI_CONVERT_MODE = "ANKI_CONVERT_MODE"

    def __init__(self, **kwargs: typing.Any):
        self.exclude_empty = kwargs.pop("exclude_empty", False)
        self.ignore_tags = set(kwargs.pop("ignore_tags_list", {}))
        self.ignore_shallow_tags = set(kwargs.pop("ignore_shallow_tags_list", {}))
        self._convert_mode = kwargs.pop("convert_mode", None)
        self.user_supplied_convert_mode = bool(self._convert_mode)

        if kwargs:
            # If we have unused options raise an exception
            ValueError(f"Unknown kwargs '{kwargs}'")

        self.node_convertor = NodeConvertor()
        self.preprocessor = Preprocessor()

    def convert(self, in_file_str: str, out_file_str: str = None) -> None:
        """Convert a single file to ANKI deck."""

        log.info(f"Converting file '{in_file_str}' to ANKI deck @ '{out_file_str}'")

        # Create paths
        in_file = pathlib.Path(in_file_str)
        if out_file_str is not None:
            assert out_file_str.endswith(AnkiConvertor.ANKI_EXT)
            out_file = pathlib.Path(out_file_str).resolve()
        else:
            out_file = in_file.with_suffix(AnkiConvertor.ANKI_EXT).resolve()
        tmp_out_file = out_file.with_suffix(".tmp.apkg").resolve()

        # Convert the org nodes into list of Notes
        cards: typing.List[genanki.Note] = []

        # Preprocess and parse the file
        preprocessed_source = self.preprocessor.preprocess_file(str(in_file))
        org_file = orgparse.loads(preprocessed_source)

        # If user did not supplied the convert mode - try to get the convert mode
        # from the org file header fall back to NORMAL mode
        if not self.user_supplied_convert_mode:
            try:
                self.convert_mode = AnkiConvertMode[
                    org_file._special_comments[self.COMMENT_ANKI_CONVERT_MODE][0].upper()
                ]
            except KeyError:
                self.convert_mode = AnkiConvertMode.NORMAL
        else:
            self.convert_mode = self._convert_mode

        self._get_cards(org_file, cards)

        # Try to set the deck name to a org file title comment
        try:
            deck_name = org_file._special_comments["TITLE"][0]
        except (KeyError, IndexError):
            deck_name = out_file.stem

        # TODO: Hash should be calculated from the cards
        deck = genanki.Deck(random.randrange(1 << 30, 1 << 31), deck_name)

        for c in cards:
            deck.add_note(c)

        package = genanki.Package(deck)
        package.media_files = self.node_convertor.media_files
        package.write_to_file(str(tmp_out_file))

        # Import and export the collection using Anki
        # This is neccessary to make mobile version work (include rendered equations)
        with utils.create_empty_anki_collection() as col:
            log.debug("Importing to tmp ANKI collection")
            imp = AnkiPackageImporter(col, str(tmp_out_file))
            imp.run()
            log.debug("Exporting from tmp ANKI collection")
            exp = AnkiPackageExporter(col)
            exp.exportInto(str(out_file))

        tmp_out_file.unlink()

    def _get_cards(
        self, tree_level: orgparse.node.OrgNode, output_list: typing.List[genanki.Note]
    ) -> None:
        for child in tree_level.children:
            # TODO: We can maybe include also cards that have children but
            # also have body text, those will have a list of all children titles?
            # TODO: This node will also contain the child node titles
            if self.should_include_node(child):
                self._process_node(child, output_list)
            self._get_cards(child, output_list)

    def should_include_node(self, node: orgparse.node.OrgNode) -> bool:
        """Determine if the node should be included."""
        return not (
            bool(self.ignore_tags.intersection(node.tags))
            or bool(self.ignore_shallow_tags.intersection(node.shallow_tags))
            or (self.exclude_empty and not node.body.strip())
        )

    def _process_node(
        self, node: orgparse.node.OrgNode, output_list: typing.List[genanki.Note]
    ) -> None:
        converted_node = self.node_convertor.convert(node, self.convert_mode)
        if converted_node:
            log.info(f"Adding node '{node.heading}' with tags '{node.tags}'")
            output_list.append(converted_node)



import contextlib

from anki.importing import AnkiPackageImporter

import orglearn.utils as utils
from orglearn.anki.anki_convertor import AnkiConvertor


@contextlib.contextmanager
def open_apkg(apkg_path):
    with utils.create_empty_anki_collection() as col:
        imp = AnkiPackageImporter(col, str(apkg_path))
        imp.run()

        yield col


def notes_iterator(col):
    for card_id in col.find_cards(""):
        yield col.getCard(card_id).note()


def test_basic(tmp_path, data_folder):
    # Initalize convertor
    c = AnkiConvertor()

    # Convert
    org_file = data_folder / "anki.org"
    out_file = tmp_path / "anki.apkg"
    c.convert(str(org_file), str(out_file))

    # Open the anki deck
    with open_apkg(str(out_file)) as col:

        notes = list(notes_iterator(col))

        expected_cards = [
            ("First node", "First body"),
            ("First node -> First First node", "First First body"),
            ("First node -> First Second node", ""),
            ("Second node", "Second body"),
            ("Second node -> Second First node", "<br />" * 4),
            ("Third node", "Third body"),
            ("Third node -> Third First node", "Third First body"),
        ]

        assert len(expected_cards) == len(notes)

        for (title, body) in expected_cards:
            assert any(title == note.fields[0] and body == note.fields[1] for note in notes)


def test_exclude_empty(tmp_path, data_folder):
    # Initalize convertor
    c = AnkiConvertor(exclude_empty=True)

    # Convert
    org_file = data_folder / "anki.org"
    out_file = tmp_path / "anki.apkg"
    c.convert(str(org_file), str(out_file))

    # Open the anki deck
    with open_apkg(str(out_file)) as col:

        notes = list(notes_iterator(col))

        expected_cards = [
            ("First node", "First body"),
            ("First node -> First First node", "First First body"),
            ("Second node", "Second body"),
            ("Third node", "Third body"),
            ("Third node -> Third First node", "Third First body"),
        ]

        assert len(expected_cards) == len(notes)

        for (title, body) in expected_cards:
            assert any(title == note.fields[0] and body == note.fields[1] for note in notes)


def test_math(tmp_path, data_folder):
    # Initalize convertor
    c = AnkiConvertor(exclude_empty=True)

    # Convert
    org_file = data_folder / "anki_math.org"
    out_file = tmp_path / "anki_math.apkg"
    c.convert(str(org_file), str(out_file))

    # Open the anki deck
    with open_apkg(str(out_file)) as col:

        notes = list(notes_iterator(col))

        expected_cards = [
            ("Mass–energy relation", "[$]E = mc^2[/$]"),
            ("Integers", "Let [$]a[/$] and [$]b[/$] be integers that are [$]\\le 0[/$]."),
        ]

        assert len(expected_cards) == len(notes)

        for (title, body) in expected_cards:
            assert any(title == note.fields[0] and body == note.fields[1] for note in notes)


def test_code(tmp_path, data_folder):
    # Initalize convertor
    c = AnkiConvertor(exclude_empty=True)

    # Convert
    org_file = data_folder / "anki_blocks.org"
    out_file = tmp_path / "anki_blocks.apkg"
    c.convert(str(org_file), str(out_file))

    # Open the anki deck
    with open_apkg(str(out_file)) as col:

        notes = list(notes_iterator(col))

        expected_cards = [
            ("Generic", "<pre><code><br />|+|<br />   -><br />   <-<br /></code></pre>"),
            (
                "Python",
                "<br /><pre><code><br />if __name__ == '__main__':<br />    print('Hello, world')<br /></code></pre><br />",
            ),
            (
                "Quote",
                "<br /><pre><code><br />Supreme excellence consists of breaking the enemy's resistance without fighting.<br /></code></pre>",
            ),
            (
                "Example",
                "<br /><pre><code><br />  table |asd|asd|<br /></code></pre>",
            ),
        ]

        assert len(expected_cards) == len(notes)

        for (title, body) in expected_cards:
            assert any(title == note.fields[0] and body == note.fields[1] for note in notes)



#!/usr/bin/env python

import typing

from setuptools import find_packages, setup


def requirements(filepath: str) -> typing.List[str]:
    """Read requirements from a filepath."""
    res = []
    with open(filepath, "r") as requirements_file:
        for line in requirements_file.readlines():
            res.append(line.strip())
    return res


with open("README.md", "r") as fh:
    long_description = fh.read()

development = ["pytest==6.2.4", "black==19.10b0", "pre-commit==2.0.1", "ipdb==0.13.7"]

setup(
    name="orglearn",
    version="1.1.5",  # TODO: Better way to handle versions
    author="Matej Kastak",
    author_email="matej.kastak@gmail.com",
    description="Tool to learn from your org-mode notes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License ",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Education",
    ],
    extras_require={
        "all": development,
        "dev": development,
    },
    include_package_data=True,
    install_requires=requirements("requirements.txt"),
    url="https://github.com/MatejKastak/orglearn",
    packages=find_packages(),
    entry_points={"console_scripts": ["orglearn = orglearn.cli:main"]},
)



import logging
import os
import sys
import tempfile
import typing

import click
import pypandoc

import orglearn.utils as utils
from orglearn.anki.anki_convertor import AnkiConvertor
from orglearn.anki.node_convertor import AnkiConvertMode
from orglearn.mind_map.backend.backends import Backends
from orglearn.mind_map.map_convertor import MapConvertor
from orglearn.preprocessor import Preprocessor

log = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option("-v", "--verbose", is_flag=True, help="Increase verbosity of orglearn.")
@click.pass_context
def main(ctx: click.Context, verbose: bool) -> None:
    """Toolbox for learning from your org notes."""

    utils.setup_loggging(logging.DEBUG if verbose else logging.WARNING)

    if ctx.invoked_subcommand is None:
        print("Hello, orglearn")
    else:
        # print('Hello, some command')
        pass


@main.command()
@click.argument("org_files", type=click.Path(exists=True), required=True, nargs=-1)
@click.option(
    "-c",
    "--conversion-mode",
    type=click.Choice(AnkiConvertMode.__members__.keys(), case_sensitive=False),
    help="Anki conversion mode.",
    default=None,
    show_default=True,
)
@click.option(
    "-e",
    "--exclude-empty",
    is_flag=True,
    help="Exclude nodes with empty body.",
    default=False,
)
@click.option(
    "-i",
    "--ignore-shallow-tag",
    "istl",
    multiple=True,
    help="Ignore specified set of shallow tags (can be specified multiple times).",
)
@click.option(
    "-I",
    "--ignore-tag",
    "itl",
    multiple=True,
    help="Ignore specified set of tags (can be specified multiple times).",
)
def anki(
    org_files: typing.Tuple[str],
    conversion_mode: typing.Optional[str],
    exclude_empty: bool,
    istl: typing.List[str],
    itl: typing.List[str],
) -> None:
    """Convert org files ORG_FILES into an anki deck."""

    # TODO: Create click.Choice for enums and return optional enum instead of this conversion
    _conversion_mode = AnkiConvertMode[conversion_mode.upper()] if conversion_mode else None

    c = AnkiConvertor(
        convert_mode=_conversion_mode,
        exclude_empty=exclude_empty,
        ignore_shallow_tags_list=istl,
        ignore_tags_list=itl,
    )

    for org_file in org_files:
        c.convert(org_file)


@main.command("map")
@click.argument("org_files", type=click.Path(exists=True), nargs=-1)
@click.option(
    "-b",
    "--backend",
    "backend_name",
    default=Backends.get_default_backend(),
    type=click.Choice(Backends.get_backends()),
    help="Backend used for conversion. Usually represents output format.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True, dir_okay=False, writable=True),
    help="Conversion output file.",
)
@click.option(
    "-i",
    "--ignore-tag",
    "itl",
    multiple=True,
    help="Ignore specified set of tags (can be specified multiple times).",
)
def map_cmd(org_files: typing.Tuple[str], backend_name: str, output: str, itl: bool) -> None:
    """Convert org files ORG_FILES into a mind map."""
    backend = Backends.make(backend_name, ignore_tags_list=itl)
    conv = MapConvertor(output, org_files, backend)


@main.command()
@click.argument("org_files", type=click.Path(exists=True), required=True, nargs=-1)
@click.option(
    "-H",
    "--heading-level",
    "heading_level",
    default=4,
    show_default=True,
    type=click.IntRange(1, 8),
    help="Set the heading level of the output pdf file.",
)
@click.option(
    "-E",
    "--exit-after-preprocessing",
    "exit_after_preprocessing",
    show_default=True,
    is_flag=True,
    type=bool,
    help="Exit after preprocessing the file and print it to stdout.",
)
def pdf(org_files: typing.Tuple[str], heading_level: int, exit_after_preprocessing: bool) -> None:
    """Convert ORG_FILES into pdf files."""

    # TODO: Test if pandoc is installed, if not report an error

    with tempfile.NamedTemporaryFile(suffix=".tex") as fp:
        fp.write(
            rb"""   \usepackage{enumitem}
    \usepackage{dsfont}
    \setlistdepth{9}

    \setlist[itemize,1]{label=$\bullet$}
    \setlist[itemize,2]{label=$\bullet$}
    \setlist[itemize,3]{label=$\bullet$}
    \setlist[itemize,4]{label=$\bullet$}
    \setlist[itemize,5]{label=$\bullet$}
    \setlist[itemize,6]{label=$\bullet$}
    \setlist[itemize,7]{label=$\bullet$}
    \setlist[itemize,8]{label=$\bullet$}
    \setlist[itemize,9]{label=$\bullet$}
    \renewlist{itemize}{itemize}{9}

    \setlist[enumerate,1]{label=$\arabic*.$}
    \setlist[enumerate,2]{label=$\alph*.$}
    \setlist[enumerate,3]{label=$\roman*.$}
    \setlist[enumerate,4]{label=$\arabic*.$}
    \setlist[enumerate,5]{label=$\alpha*$}
    \setlist[enumerate,6]{label=$\roman*.$}
    \setlist[enumerate,7]{label=$\arabic*.$}
    \setlist[enumerate,8]{label=$\alph*.$}
    \setlist[enumerate,9]{label=$\roman*.$}
    \renewlist{enumerate}{enumerate}{9}
"""
        )

        fp.flush()

        old_working_directory = os.getcwd()

        for file_path in org_files:

            input_content = ""
            input_content += r"#+LATEX_HEADER: \setlength\parindent{0pt}"
            input_content += "\n"
            input_content += "#+OPTIONS: tags:nil"
            input_content += "\n"
            input_content += "#+OPTIONS: ^:nil"
            input_content += "\n"

            if heading_level:
                input_content += "#+OPTIONS: H:{}".format(heading_level)
                input_content += "\n"

            input_content += Preprocessor().preprocess_file(file_path)

            if exit_after_preprocessing:
                print(input_content)
                sys.exit(0)

            # Change the compilation context to the file directory
            # this is needed in order to include images from the paths relative to org file
            os.chdir(os.path.dirname(os.path.abspath(file_path)))

            of = os.path.splitext(os.path.basename(file_path))[0] + ".pdf"

            log.info("Calling pandoc to generate PDF")
            try:
                pypandoc.convert_text(
                    input_content,
                    "pdf",
                    "org",
                    outputfile=of,
                    extra_args=[
                        "--pdf-engine=xelatex",
                        # TODO: Add an option to set the documentclass
                        "-V",
                        "documentclass=report",
                        "-V",
                        "block-headings",
                        "--toc",
                        "-N",
                        "-V",
                        "geometry:top=2.5cm, bottom=2.5cm, left=4cm, right=4cm",
                        "-H",
                        fp.name,
                    ],
                )
            except RuntimeError as e:
                error_text = bytes(str(e), "utf-8").decode("unicode_escape")
                log.error(f"Failed to compile pdf:\n{error_text}")
                sys.exit(1)
            log.info("PDF created")

        # Restore the old working directory
        os.chdir(old_working_directory)


@main.command()
@click.argument("org_file", type=click.Path(exists=True), required=True, nargs=1)
def preprocess(org_file: str) -> None:
    """Preprocess ORG_FILES and print them to stdout.

    NOTE: This is used for debugging purposes.
    """
    print(Preprocessor().preprocess_file(org_file))



import abc

# TODO(mato): Consider moving changing this to Convertor base class
# it will have similiar interface, but it will output to the specified
# stream
class Backend(abc.ABC):
    @abc.abstractmethod
    def convert(self, tree, o_stream, **kwargs):
        """Convert org tree with backend.
        Keyword Arguments:
        tree     -- Org tree input.
        o_stream -- Write output into the stream.
        **kwargs -- Optional arguments for backend.
        """
        ...

    def get_ext(self) -> str:
        raise NotImplemented()
